// Global variables

var nodeIdMap = {};

// Map of kernel root node id2 -> bool
var initializedKernels = {};

// OFF: the node itself cannot be expanded
// ONE: only one sibling can be expanded at a time
// PASS: the node itself doesn't respong to clicks but passes it to parent
// ALL: default, everything can be expanded 
var ExpEnum = { OFF: "off", ONE: "one", PASS: "pass", ALL: "all" };

var maxID2 = 0;

// Main function
function
startGraph( data )
{
  // 0. Reset all data structures and remove all DOM elements
  reset();

  // 1. Prep the data in various ways
  if ( initializedKernels[ data.id2 ] != true ) {
    // Mark the kernel as initialized
    initializedKernels[ data.id2 ] = true;

    // 1f. For each node, remember all children (hidden or not) in the 
    // allchildren member variable
    addAllChildren( data );

    // 1b. Add shape-related attribute to each node based on its type
    addShapeAttributes( data );

    // 1c. For each node, decide how many of the children can be expanded at
    // a time 
    addExpandAttribute( data );

    // 1d. For each node, decide whether children should be hidden or visible
    // by default
    // Also set the collapsed/expanded attribute
    addChildrenVisibility( data );

    // 1e. For each node, add a pointer to its parent
    // This is used to collapse one child if another is expanded for nodes that
    // require this functionality
    addParent( data );

    // 1f. Convert details to an array for easier use with displaying the
    // details table
    prepareDetails( data );
  }
  
  // 1.b Find max id2. This will be used to assign unique ids to edges
  findMaxID2( data );
 
  // 1a. Create a map of node id -> node object
  // This will be used to create the link objects
  nodeIdMap = populateNodeIdMap( data );
 
////  console.log( "Final data from main.js: " );
////  console.log( data );
  
  // 2. Create the diagram
  var flowTree = d3.custom.chart.flow();

  d3.select('#flowtree')
    .datum(data)
    .call(flowTree)
    .on("mouseover.customHover", function() {
        //    console.log("customHover");
     } )
    .on("resize.custom", function() {
      //    console.log("resize.custom");
     } );

  /// Functions
  function
  reset()
  {
    nodeIdMap = {};
    maxID2 = 0;

    var node = d3.select( '#flowtree' );
    node.each( function() {
      while ( this.lastChild ) this.removeChild( this.lastChild );
    } );
  }
}

function
addAllChildren( node )
{
  assert( !node._children );
  var children = [];
  if ( node.children ) children = children.concat( node.children );
  node.allchildren = children;
  
  children.forEach( addAllChildren );
}

function
findMaxID2( node )
{
  if ( node.id2 > maxID2 ) maxID2 = node.id2;
  node.allchildren.forEach( findMaxID2 );
}

function
populateNodeIdMap( root )
{
  var stack = [ root ],
      node,
      nodeMap = {};

  while( ( node = stack.pop() ) != null ) {
    nodeMap[ node.id2 ] = node;
    var children = node.allchildren;   
    
    var n = children.length;
    while ( --n >= 0 ) {
      stack.push( children[ n ] );
    }
  }

  return nodeMap;
}

function
addShapeAttributes( node )
{
  // Set the node shape and intrinsic width/height
  switch( node.type ) {
    case "merge": 
    case "branch":
    case "port":
      {
        node.shape = "CIRC";
        // TODO fix these to be 20 once I've moved them
        // to the basic block borders
        // In the case of the circle iw=ih=diameter
        node.iw = 8;
        node.ih = 8;
      }
      break;

    case "kernel":
    case "container":
      {
        node.shape = "RECT";
        node.iw = 120;
        node.ih = 60;
      }
      break;

    case "bb":
    case "memtype":
    case "memsys":
    case "bank":
      {
        node.shape = "RECT";
        node.iw = 100;
        node.ih = 40;
      }
      break;

    case "inst":
      {
        node.shape = "RECT";
        node.iw = 80;
        node.ih = 20;
      }
      break;

    default:
      // TODO
      break;
  }

  node.allchildren.forEach( addShapeAttributes );
}

// OFF: the node itself cannot be expanded
// ONE: only one sibling can be expanded at a time
// PASS: the node itself doesn't respond to clicks but passes it to parent
// ALL: default, everything can be expanded 
// var ExpEnum = { OFF: "off", ONE: "one", PASS: "pass", ALL: "all" };
function
addExpandAttribute( node )
{
  // Set node.exp to one of the above expand options 
  switch( node.type ) {
    case "memtype":
    case "memsys":
      node.exp = ExpEnum.ONE;
      break;

    case "bank":
      node.exp = ExpEnum.PASS;
      break;

    case "container":
    case "inst":
    case "port":
      node.exp = ExpEnum.OFF;
      break;

    default:
      node.exp = ExpEnum.ALL;
      break;
  }

  node.allchildren.forEach( addExpandAttribute );
}

function
addChildrenVisibility( node )
{
  assert( !node._children );

  switch( node.type ) {
    // Visible by default 
    case "container":
    case "bank":
    {
      // Do nothing
      // Visible by default
      node.children = node.allchildren;
      node._children = null;
      node.expanded = true;
    }
    break;

    // Hidden by default
    default:
    {
      node._children = node.allchildren;
      node.children = null;
      node.expanded = false;
    } 
    break;
  }

  node.allchildren.forEach( addChildrenVisibility );
}

function
addParent( node )
{
  // If the parent hasn't already been set, set it to null (for the root)
  if ( !node.parent ) node.parent = null;
  
  node.allchildren.forEach( function( child ) {
    child.parent = node;
    addParent( child );
  } );
}

function
prepareDetails( node )
{
  if ( node.details ) {
    var detailsArray = [];
    var detailsMap = node.details;
    for ( var key in detailsMap ) {
      if ( detailsMap.hasOwnProperty( key ) ) {
        var value = detailsMap[ key ];
        detailsArray.push( { first: key, second: value } );
      }
    }
    node.details = detailsArray;
  }
  
  node.allchildren.forEach( prepareDetails );
}
