
# Required packages
package require ::quartus::project
package require ::quartus::report
package require ::quartus::flow


# Utility functions

# Return values: [retval panel_id row_index]
#   panel_id and row_index are only valid if the query is successful
# retval: 
#    0: success
#   -1: not found
#   -2: panel not found (could be report not loaded)
#   -3: no rows found in panel
#   -4: multiple matches found
proc find_report_panel_row { panel_name col_index string_op string_pattern } {
    if {[catch {get_report_panel_id $panel_name} panel_id] || $panel_id == -1} {
        return -2;
    }

    if {[catch {get_number_of_rows -id $panel_id} num_rows] || $num_rows == -1} {
        return -3;
    }

    # Search for row match.
    set found 0
    set row_index -1;

    for {set r 1} {$r < $num_rows} {incr r} {
        if {[catch {get_report_panel_data -id $panel_id -row $r -col $col_index} value] == 0} {
            if {[string $string_op $string_pattern $value]} {
                if {$found == 0} {
                    # If multiple rows match, return the first
                    set row_index $r
                }
                incr found
            }

        }
    }

    if {$found > 1} {return [list -4 $panel_id $row_index]}
    if {$row_index == -1} {return -1}

    return [list 0 $panel_id $row_index]
}


# get_fmax_from_report: Determines the fmax for the given clock. The fmax value returned
# will meet all timing requirements (setup, hold, recovery, removal, minimum pulse width)
# across all corners.  The return value is a 2-element list consisting of the
# fmax and clk name
proc get_fmax_from_report { clkname required } {
    # Find the clock period.
    set result [find_report_panel_row "TimeQuest Timing Analyzer||Clocks" 0 match $clkname]
    set retval [lindex $result 0]

    if {$retval < 0} { 
        post_message -type warning "Could not find clock: $clkname."
        return [list -1 $clkname]
    }

    # Update clock name to full clock name ($clkname as passed in may contain wildcards).
    set panel_id [lindex $result 1]
    set row_index [lindex $result 2]
    set clkname [get_report_panel_data -id $panel_id -row $row_index -col 0]
    set clk_period [get_report_panel_data -id $panel_id -row $row_index -col 2]

    post_message "Clock $clkname"
    post_message "  Period: $clk_period"

    # Determine the most negative slack across all relevant timing metrics (setup, recovery, minimum pulse width)
    # and across all timing corners. Hold and removal metrics are not taken into account
    # because their slack values are independent on the clock period (for kernel clocks at least).
    #
    # Paths that involve both a posedge and negedge of the kernel clocks are not handled properly (slack
    # adjustment needs to be doubled).
    set timing_metrics [list "Setup" "Recovery" "Minimum Pulse Width"]
    set timing_metric_colindex [list 1 3 5 ]
    set timing_metric_required [list 1 0 0]
    set wc_slack $clk_period
    set has_slack 0
    set fmax_from_summary 5000.0

    # Find the "Fmax Summary" numbers reported in Quartus.  This may not
    # account for clock transfers but it does account for pos-to-neg edge same
    # clock transfers.  Whatever we calculate should be less than this.
    set fmax_panel_name "TimeQuest Timing Analyzer||* Model||* Model Fmax Summary"
    foreach panel_name [get_report_panel_names] {
      if {[string match $fmax_panel_name $panel_name] == 1} {
        set result [find_report_panel_row $panel_name 2 equal $clkname]
        set retval [lindex $result 0]
        if {$retval == 0} {
          set restricted_fmax_field [get_report_panel_data -id [lindex $result 1] -row [lindex $result 2] -col 1]
          regexp {([0-9\.]+)} $restricted_fmax_field restricted_fmax 
          if {$restricted_fmax < $fmax_from_summary} {
            set fmax_from_summary $restricted_fmax
          }
        }
      }
    }
    post_message "  Restricted Fmax from STA: $fmax_from_summary"
    set fmax $fmax_from_summary

#   # Find the worst case slack across all corners and metrics
#   foreach metric $timing_metrics metric_required $timing_metric_required col_ndx $timing_metric_colindex {
#     set panel_name "TimeQuest Timing Analyzer||Multicorner Timing Analysis Summary"
#     set panel_id [get_report_panel_id $panel_name] 
#     set result [find_report_panel_row $panel_name 0 equal " $clkname"]
#     set retval [lindex $result 0]
#
#     if {$retval == -1} { 
#       if {$required == 1 && $metric_required == 1} {
#         error "Error: Could not find clock: $clkname" 
#       }
#     } elseif {$retval < 0 && $retval != -4 } { 
#       error "Error: Failed search for clock $clkname (error $retval)"
#     }
#
#     if {$retval == 0 || $retval == -4} {
#       set slack [get_report_panel_data -id [lindex $result 1] -row [lindex $result 2] -col $col_ndx ]
#
#       post_message "    $metric slack: $slack"
#       set has_slack 1
#
#       # Keep track of the most negative slack.
#       if {$slack < $wc_slack} {
#         set wc_slack $slack
#         set wc_metric $metric
#         }
#     }
#   }
#
#   if {$has_slack == 1} {
#       # Adjust the clock period to meet the worst-case slack requirement.
#       set clk_period [expr $clk_period - $wc_slack]
#       post_message "  Adjusted period: $clk_period ([format %+0.3f [expr -$wc_slack]], $wc_metric)"
#
#       # Compute fmax from clock period. Clock period is in nanoseconds and the
#       # fmax number should be in MHz.
#       set fmax [expr 1000 / $clk_period]
#
#       if {$fmax_from_summary < $fmax} {
#           post_message "  Restricted Fmax from STA is lower than $fmax, using it instead."
#           set fmax $fmax_from_summary
#       }
#
#       # Truncate to two decimal places. Truncate (not round to nearest) to avoid the
#       # very small chance of going over the clock period when doing the computation.
#       set fmax [expr floor($fmax * 100) / 100]
#       post_message "  Fmax: $fmax"
#   } else {
#       post_message -type warning "No slack found for clock $clkname - assuming 10 GHz."
#       set fmax 10000
#   }

    return [list $fmax $clkname]
}


proc get_entity_resources_incl_subhier { fit_rpt_resource } {
    regexp {([0-9.]+)\s+\([0-9.]+\)} $fit_rpt_resource match usage

    return $usage
}

##############################################################################
##############################       MAIN        #############################
##############################################################################
set is_pro [sw_edition_is -pro]

set project_name [lindex $quartus(args) 1]
set project_rev  [lindex $quartus(args) 2]

project_open $project_name -revision $project_rev

load_report $project_name

# Process arguments.
set fmax1 [ lindex [get_fmax_from_report "clock"   1] 0]
set fmax2 [ lindex [get_fmax_from_report "clock2x" 1] 0]

# Get the project's target family
set family [get_global_assignment -name FAMILY]

# A little report
# Since we might be paritioning the components and registering them, we need to get the resources from the "by Entity" summary
set by_entity_panel_str "Fitter||Resource Section||Fitter Resource Utilization by Entity"
if { $is_pro } {
  set by_entity_panel_str "Fitter||Place Stage||Fitter Resource Utilization by Entity"
}
set by_entity_panel_id [get_report_panel_id $by_entity_panel_str]
set by_entity_num_rows [get_number_of_rows -id $by_entity_panel_id]

# Setup the columns to be analyzed based on family
set alms_col       1
set registers_col  7
set dsp_col       11
set ram_col       10
if { [string compare $family "MAX10"] == 0 } {
  set alms_col      1
  set registers_col 2
  set dsp_col       7
  set ram_col       5
}

set alms      0
set registers 0
set dsp       0
set mem_bit   0
set rams      0

for {set r 0} {$r < $by_entity_num_rows} {incr r} {
    if {[catch {get_report_panel_data -id $by_entity_panel_id -row $r -col 0} entity] == 0} {
        if {[regexp {hls_component_dut_inst_[0-9]+} $entity match] } {
            set comp_alms      [get_entity_resources_incl_subhier [get_report_panel_data -id $by_entity_panel_id -row $r -col $alms_col     ] ]
            set comp_registers [get_entity_resources_incl_subhier [get_report_panel_data -id $by_entity_panel_id -row $r -col $registers_col] ]
            set comp_dsp                                          [get_report_panel_data -id $by_entity_panel_id -row $r -col $dsp_col      ]
            set comp_rams                                         [get_report_panel_data -id $by_entity_panel_id -row $r -col $ram_col      ]

            set alms      [expr $alms      + $comp_alms     ]
            set registers [expr $registers + $comp_registers]
            set dsp       [expr $dsp       + $comp_dsp      ]
            set rams      [expr $rams      + $comp_rams     ]
        }
    }
}

set logic_name "ALMs"
if { [string compare $family "MAX10"] == 0 } {
  set logic_name "LCs"
}

set outfile   [open "qii_compile_report.txt" w]
puts $outfile "$logic_name: $alms"
puts $outfile "Registers: $registers"
puts $outfile "DSP blocks: $dsp"
puts $outfile "RAM blocks: $rams"
puts $outfile "1x clock fmax: $fmax1 MHz"
if {$fmax2 != -1} {
    puts $outfile "2x clock fmax: $fmax2 MHz"
}

# Highest non-global fanout signal
#set result [find_report_panel_row "Fitter||Resource Section||Fitter Resource Usage Summary" 0 equal "Highest non-global fan-out"]
#if {[lindex $result 0] < 0} {error "Error: Could not find highest non-global fan-out (error $retval)"}
#set high_fanout_signal_fanout_count [get_report_panel_data -id [lindex $result 1] -row [lindex $result 2] -col 1]

#puts $outfile "Highest non-global fanout: $high_fanout_signal_fanout_count"

close $outfile
# End little report

unload_report

